class WeldError(Exception): ...


class InvalidMcmeta(WeldError):
    def __init__(self, pack: str, contents: str):
        super().__init__(f"Pack {pack}'s mcmeta file is not valid JSON:\n{contents}")
        self.pack = pack
        self.contents = contents



import time
from concurrent.futures import ThreadPoolExecutor
from importlib import resources
from pathlib import Path
from zipfile import ZipFile

import streamlit as st
import yaml
from streamlit.delta_generator import DeltaGenerator
from streamlit_extras.stateful_button import button as toggle_button
from streamlit_extras.streaming_write import write as streaming_write

from smithed import weld

from .log_helpers import init_logger
from .models import Columns, WebApp

icon = "https://github.com/Smithed-MC/smithed-python/blob/main/smithed/weld/resources/icon.png?raw=true"

webapp = WebApp.parse_obj(
    yaml.safe_load(
        (resources.files("smithed") / "weld/resources/webapp.yaml").read_text()
    )
)


def weld_packs(packs: list[ZipFile], make_fabric_mod: bool) -> Path | None:
    """Welds a list of zip files. Outputs a path"""

    with weld.run_weld(packs, as_fabric_mod=make_fabric_mod) as ctx:
        if make_fabric_mod:
            with ZipFile("output.jar", "w") as jar:
                ctx.data.dump(jar)
                ctx.assets.dump(jar)
                output = Path(str(jar.filename))

        # TODO: figure out what to do if user uploads both data and resource packs
        # prolly merge and download UNZIP ME or something
        if ctx.data:
            output = ctx.data.save(path="output", zipped=True, overwrite=True)

        if ctx.assets:
            output = ctx.assets.save(path="output", zipped=True, overwrite=True)

        return output  # type: ignore ikkk TODO


def build_packs(
    packs: list[ZipFile], cols: Columns, make_fabric_mod: bool = False
) -> Path | None:
    path = None

    with st.status(f"Welding {len(packs)} packs!", expanded=False) as status:
        try:
            t0 = time.perf_counter()
            stream = init_logger()

            with ThreadPoolExecutor() as executor:
                future = executor.submit(
                    weld_packs, packs=packs, make_fabric_mod=make_fabric_mod
                )
                while not future.done():
                    time.sleep(0.1)
                    streaming_write(val := stream.getvalue())
                    print(val)
                streaming_write(val := stream.getvalue())
                print(val)

                path = future.result()

            t1 = time.perf_counter()
            status.update(
                label=f"Merged in :green[{t1 - t0: 0.3f}s]. Click to see log."
            )

            return path

        except Exception as exc:
            status.update(label=":red[Error occured. Click to reveal error.]")
            status.error(f"# `{exc.__class__.__name__}`\n{exc.args[0]}")

    return path


def upload_flow(ui: DeltaGenerator):
    progress = ui.container()
    raw_packs = ui.file_uploader("Upload packs", accept_multiple_files=True, type="zip")
    packs = [ZipFile(pack) for pack in raw_packs] if raw_packs else []

    cols = Columns(*ui.columns(3))
    with cols.middle:
        if fabric_mod := toggle_button("Turn into Fabric Mod", key="toggle"):
            ui.warning(webapp.fabric)

    path = None
    if cols.left.button("Build Packs", disabled=not packs, key="build") and packs:
        path = build_packs(packs, cols, make_fabric_mod=fabric_mod)
    else:
        cols.right.button("Download Welded Packs", disabled=True, key="download")

    if not packs:
        progress.info("Waiting for uploaded packs")
    elif path is None:
        progress.success("Ready to Build!")

    if path is not None:
        with path.open(mode="rb") as file:
            cols.right.download_button(
                "Download Welded Packs",
                file,
                file_name=("welded-pack.zip" if not fabric_mod else "welded-mod.jar"),
            )


def main():
    st.set_page_config(
        page_title="Weld",
        page_icon=icon,
        layout="wide",
    )

    st.sidebar.title("Special Merging Rules")
    st.sidebar.write(webapp.conflicts)
    st.sidebar.write("----")
    st.sidebar.write("See [docs](https://wiki.smithed.dev/weld) for more info!")
    st.sidebar.write("----")
    st.sidebar.warning(webapp.warn)

    st.write(webapp.title.format(version=weld.__version__), unsafe_allow_html=True)

    st.write(webapp.intro, unsafe_allow_html=True)

    upload_flow(st.container())

    st.write("---")

    col1, col2 = st.columns(2)
    col1.markdown(webapp.footer.left, unsafe_allow_html=True)
    col2.markdown(webapp.footer.right, unsafe_allow_html=True)


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from typing import NamedTuple

from pydantic.v1 import BaseModel
from streamlit.delta_generator import DeltaGenerator


class WebApp(BaseModel):
    class Footer(BaseModel):
        left: str
        right: str

    title: str
    conflicts: str
    intro: str
    warn: str
    fabric: str
    footer: Footer


class Columns(NamedTuple):
    left: DeltaGenerator
    middle: DeltaGenerator
    right: DeltaGenerator
    middle: DeltaGenerator
    right: DeltaGenerator


@dataclass
class WeldSettings:
    output_name: str = "welded-pack"
    description: str = "This pack contains multiple packs welded together"
    make_fabric_mod: bool = False



"""
This file is super scrappy and needs to be rewritten but I can't be bothered right now.
"""

import logging
from dataclasses import dataclass
from typing import Any, Iterator, NamedTuple

from tokenstream import Token, TokenStream

from smithed.type import JsonDict, JsonList, JsonType

from .errors import ParsingError

logger = logging.getLogger("weld")


@dataclass
class Filter:
    key: str
    value: Any
    type: str = "filter"


class TraverseResult(NamedTuple):
    parent: JsonDict | JsonList
    current: JsonType
    key: Any


def parse(raw: str, convert_index: bool):
    stream = TokenStream(raw)
    return handle_stream(stream, convert_index)


def handle_stream(stream: TokenStream, convert_index: bool) -> Iterator[Token | Filter]:
    with stream.syntax(
        key=r"[A-Z_a-z]+",
        string=r'"\w*"',
        separator=r"\.",
        bracket=r"\[|\]",
        curly=r"\{|\}",
        number=r"\d+",
        colon=r": *",
    ):
        for token in stream.collect():
            match token:
                case Token(type="key"):
                    yield token

                case Token(type="bracket", value="["):
                    match stream.expect_any("number", ("curly", "{")):
                        case Token(type="number", value=value) as token:
                            if convert_index:
                                yield Filter(key="_index", value=int(value))
                            else:
                                yield token
                        case Token(type="curly"):
                            yield parse_filter(stream)
                            stream.expect(("curly", "}"))

                        case Token() as token:
                            raise ParsingError(
                                f"Unexpected token {token} reached."
                                " Expected number or filter subexpression."
                            )

                    stream.expect(("bracket", "]"))

                case Token(type="separator") as token:
                    yield token

                case Token() as token:
                    raise ParsingError(
                        f"Unexpected token {token} reached."
                        " Expected key, bracket, or separator."
                    )


def parse_filter(stream: TokenStream):
    match stream.expect_any("key", "string"):
        case Token(type="key", value=value):
            key = value
        case Token(type="string", value=value):
            key = value[1:-1]
        case Token() as token:
            raise ValueError(f"Expected key or string. Found {token}")

    stream.expect("colon")
    return Filter(key=key, value=stream.expect_any("string", "number").value)


def traverse(obj: JsonDict, path: str, convert_index: bool = False):
    current = obj
    parent = current
    parser = parse(path, convert_index)
    last_token = None

    for token in parser:
        match token:
            case Token(type="key", value=value):
                if not isinstance(current, dict):
                    raise ValueError
                parent = current
                current = current.setdefault(value, {})
                last_token = token

            case Token(type="number", value=index):
                if not isinstance(current, list):
                    raise ValueError
                elif len(current) <= int(index):
                    raise IndexError
                parent = current
                current = current[int(index)]
                last_token = token

            case Filter(key=key, value=value):
                if not isinstance(current, list):
                    if last_token is None:
                        raise ValueError
                    if not isinstance(parent, dict):
                        raise ValueError
                    current = []
                    parent[last_token.value] = current  # type: ignore

                for item in current:
                    if not isinstance(item, dict):
                        raise ValueError
                    if key in item and item[key] == value:
                        parent = current
                        current = item
                        break
                else:
                    current.append({key: value})
                    parent = current
                    current = current[-1]
                last_token = token

    return TraverseResult(
        parent, current, last_token.value if last_token is not None else None
    )


def get(obj: JsonDict, path: str, convert_index: bool = False):
    return traverse(obj, path, convert_index).current


def append(obj: JsonDict, path: str, value: JsonType):
    parent, current, key = traverse(obj, path, True)
    if isinstance(current, list):
        current.append(value)
    elif not current:
        parent[key] = [value]


def prepend(obj: JsonDict, path: str, value: JsonType):
    parent, current, key = traverse(obj, path, True)
    if isinstance(current, list):
        current.insert(0, value)
    elif not current:
        parent[key] = [value]


def insert(obj: JsonDict, path: str, index: int, value: JsonType):
    parent, current, key = traverse(obj, path, True)
    if isinstance(current, list):
        current.insert(index, value)
    elif not current:
        parent[key] = [value]


def remove(obj: JsonDict, path: str):
    parent, _, key = traverse(obj, path, True)
    if isinstance(parent, list):
        del parent[int(key)]
    else:
        del parent[key]


def merge(obj: JsonDict, path: str, value: JsonType):
    parent, current, key = traverse(obj, path, True)

    def _merge(
        parent: JsonDict | JsonList, original: JsonType, to_merge: JsonType, key: str
    ):
        if isinstance(original, dict) and isinstance(to_merge, dict):
            for new_key, val in to_merge.items():
                if og_val := original.get(new_key, False):
                    parent = original
                    _merge(parent, og_val, val, new_key)
                else:
                    original[new_key] = val

        elif isinstance(original, list) and isinstance(to_merge, list):
            original.extend(to_merge)

        else:
            # logger.warn("Deprecated. Use `smithed:replace` instead.")
            parent[key] = to_merge  # TODO: this might error !!

    _merge(parent, current, value, key)


def replace(obj: JsonDict, path: str, value: JsonType):
    parent, _, key = traverse(obj, path, True)
    parent[key] = value
    parent[key] = value
